;###############################################################################
/* Selector for LilBuilder

*/
;###############################################################################

;create selector gui
Selector_Create() {                     
   Gui, 1:+LastFound -Caption +Resize +ToolWindow +LabelSelector_ ;AlwaysOnTop
   Gui, 1:Color, FFFF
   WinSet, Transparent, 100
   Selector_vHWND(WinExist())
}

;###############################################################################
;##################   Functions for otherwise global vars                                                    
;###############################################################################

Selector_vHWND(NewHWND=""){            ;selectors own gui HWND
  static HWND
  If NewHWND
     HWND = %NewHWND%
  Else Return %HWND%
}

Selector_vCtrlHWND(NewHWND=""){       ;currently selected control
  static HWND
  If NewHWND = No
     HWND =
  Else If NewHWND
     HWND = %NewHWND%
  Else Return %HWND%
}

Selector_vGuiHWND(NewHWND=""){        ;currectly "active" (selected) user gui
  static HWND
  If NewHWND = No                       ;there should normally always be one, toolbox relies on it
     HWND =
  Else If NewHWND
     HWND = %NewHWND%
  Else Return %HWND%
}

Selector_vMultiHWND(NewHWND=""){      ;list of selected controls in multi selection
  static HWND
  If NewHWND = No
     HWND =
  Else If NewHWND {
     HWND .= "|" NewHWND
     If InStr(HWND,"|") = 1
        StringTrimLeft, HWND, HWND, 1 
  }Else Return %HWND%
}

;store and retrieve position and size of selector (mainly for multi selection)
Selector_vPosSize(ByRef X, ByRef Y, ByRef W, ByRef H){
  static sX, sY, sW, sH
  If X
    sX := X, sY := Y, sW := W, sH := H,
  Else X := sX, Y := sY, W := sW, H := sH,
}

;###############################################################################
;##################   Position selector to something                                                       
;###############################################################################

;set position and size of selector gui
Selector_Set2PosSize(X, Y, W = 6, H = 6){
  w -= 6, h -= 6
  Gui, 1:Show, x%x% y%y% w%w% h%h% NA
}

;set selector to the dimensions of a control
; called when:
;   - selector needs to update itself after selector got resized/moved with GRID in single selection mode
;   - after a new control gets added via toolbox
;   - a single control gets selected with a click
;   - a single control gets selected with frame selection
Selector_Set2Ctrl(){
   WinGetPos, ux, uy,,, % "ahk_id " Selector_vGuiHWND()           ;get pos of user gui
   ControlGetPos, sx,sy,sw,sh,, % "ahk_id " Selector_vCtrlHWND()  ;get controls  pos/size
   Selector_Set2PosSize(sx + ux, sy + uy, sw, sh)                      ;show selector at pos/size of control 
}

;set selector to the dimensions of a multiselection
; called when:
;   - a single control gets add to multiselection
;   - multiple controls get selected with frame selection
Selector_Set2MultiSel(){
  MultiSelCtrlHWNDs := Selector_vMultiHWND()
  maxx := 0 ,            maxy := 0                   ;set extreme start coordinates for surrounding box
  minx := A_ScreenWidth, miny := A_ScreenHeight   
  Loop, Parse, MultiSelCtrlHWNDs, |                  ;go through all controls in selection
    {
      ControlGetPos, x,y,w,h,, ahk_id %A_LoopField%  ;get control pos/size
      maxx := maxx > x + w ? maxx : x + w            ;calculate surrounding box coordinates
      maxy := maxy > y + h ? maxy : y + h
      minx := minx < x ? minx : x
      miny := miny < y ? miny : y
    }
  sW := maxx - minx , sH := maxy - miny              ;calculate size of surrounding box
  WinGetPos, ux, uy,,, % "ahk_id " Selector_vGuiHWND()   ;get position of user gui
  sX := minx + ux , sY := miny + uy                        ;calculate global selector coordinates 
  Selector_Set2PosSize(sX, sY, sW, sH)                        ;set new pos/size for selector
  Selector_vPosSize(sX, sY, sW, sH)                    ;store new pos/size
}

;snap selector position to grid
Selector_Set2Grid(){
  GuiHWND := Selector_vGuiHWND()
  WinGetPos, ux, uy, , , ahk_id %GuiHWND%                  ;get pos of user gui
  WinGetPos, sx, sy ,sw, sh, % "ahk_id " Selector_vHWND()  ;get current pos/size of selector

  If Gui(GuiHWND,"Toolwindow"){
    ;window frame height/width of a toolwindow gui
    SysGet, XFrame,  7   ;SM_CXFIXEDFRAME =  7   ;width of frame
    SysGet, YFrame,  8   ;SM_CYFIXEDFRAME =  8   ;height of frame
    SysGet, HCaption, 51 ;SM_CYSMCAPTION  = 51   ;height of caption, the whole title height is: SM_CYSMCAPTION + SM_CYFIXEDFRAME
  }Else{
    ;window frame height/width of a regular gui
    SysGet, XFrame, 32   ;SM_CXSIZEFRAME = 32    ;width of frame
    SysGet, YFrame, 33   ;SM_CYSIZEFRAME = 33    ;height of frame
    SysGet, HCaption, 4  ;SM_CYCAPTION   =  4    ;height of caption, the whole title height is: SM_CYCAPTION + SM_CYSIZEFRAME
  }

  rx := sx - ux - XFrame                                   ;calculate relative position
  ry := sy - uy - YFrame - HCaption                                 
  GRID := Grid()                                                   ;get GRID size
  dx := Mod(rx, GRID), dy := Mod(ry, GRID)                         ;calculate delta x/y for snap to grid
  HalfGrid := Grid / 2                                             ;snap to closes grid
  dx -= (dx > HalfGrid ? GRID : 0), dy -= (dy > HalfGrid ? GRID : 0)                    
  Selector_Set2PosSize(sX - dx, sY - dy, sW, sH)                      ;set selector to new position
}

;move selector by increments with hotkeys up/down/left/right
Selector_SetMoveIncrement:
  If (Selector_vMultiHWND() OR Selector_vCtrlHWND()){
    Selector_SetMoveIncrement(A_ThisHotkey)
    SetTimer, Selector_MoveCtrl2Selector, -50                            ;update control(s)
    Grid_Draw()                                                   ;redraw grid
  }
Return
Selector_SetMoveIncrement(Direction){
  WinGetPos, sx,sy,sw,sh, % "ahk_id " Selector_vHWND()           ;get current pos/size of selector
  If Direction = Up                                             ;increment position
      sy--  
  Else If Direction = Down
      sy++
  Else If Direction = Left
      sx--
  Else If Direction = Right
      sx++
  Selector_Set2PosSize(sX, sY, sW, sH)                             ;position selector
}

;###############################################################################
;##################   Click + LCTRL to add control to selection                                                       
;###############################################################################

;add a control to a multi selection
; called when:
;   - a single control gets selected with LCtrl down
Selector_AddCtrlToSelection(CtrlHwnd){
  If (SingelSelCrtlHWND := Selector_vCtrlHWND()) { ;when a single control was selected previously
    Selector_vMultiHWND(SingelSelCrtlHWND)              ;add control from single selection to multiselection as first control
    Selector_vCtrlHWND("No")                            ;remove control from single selection
  }
  Selector_vMultiHWND(CtrlHwnd)                    ;add current control to multiselection
  Selector_Set2MultiSel()                            ;change pos/size of selector to cover all controls 
}

;###############################################################################
;##################   Draw a frame for multi selection                                                       
;###############################################################################

;drag a frame for selection
;called with timer when:
; - user clicked with LButton on gui without selecting a control
Selector_DragFrame:
  Selector_DragFrameDraw()
  If !GetKeyState("LButton")        ;end of drag on release of LButton
    Selector_DragFrameEnd()
Return
;draw a frame between start and end coords 
Selector_DragFrameDraw(X="", Y="", Action=""){
  static sX, sY
  If X                                            ;initialized start coordinates
    Return sX := X, sY := Y
  MouseGetPos, eX, eY                             ;get current mouse pos
  X := sX > eX ? eX : sX, Y := sY > eY ? eY : sY  ;calculate frame position
  W := Abs(sX - eX), H := Abs(sY - eY)            ;calculate size
  Selector_Set2PosSize(X, Y, W, H)                   ;show frame
  If Action = End                                 ;for last call store new position
    Selector_vPosSize(x,y,w,h)
}
;end of frame drag (when user releases LButton)
Selector_DragFrameEnd(){
  SetTimer, Selector_DragFrame, Off
  Selector_DragFrameDraw("", "", "End")
  Selector_SelectCtrlsByFramePosSize()   ;select controls based on frame pos/size
  If Selector_vCtrlHWND()                ;update selector to new single selection
    Selector_Set2Ctrl()
  Else If Selector_vMultiHWND()          ;update selector to new multi selection
    Selector_Set2MultiSel()
  Else
    Selector_Set2PosSize(-10, -10)
  Messages_SetHandlers( True )
  Grid_Draw()
  ToolTip_ShowTips4Pos()
  Sleep, 20
  Selector_Size(True)
}

;select controls based on frame pos/size
; called when:
;   - frame was dragged to select controls
Selector_SelectCtrlsByFramePosSize(){
  GuiHWND := Selector_vGuiHWND()                          ;which Gui
  WinGetPos, ux, uy,,, % "ahk_id " Selector_vGuiHWND()    ;get pos of user gui
  Selector_vPosSize(BoxX, BoxY, BoxW, BoxH)               ;get frame dimentions
  BoxX -= ux, BoxY -= uy                                  ;relative position
  BoxW += BoxX, BoxH += BoxY
  ListOfCtrlHWND := Gui_GetAllCtrlHWNDWTabActive(GuiHWND) ;get only controls visible on tab pages and gui
  Loop, Parse, ListOfCtrlHWND, |                         ;go through all controls on gui
    {                                           
      ControlGetPos, cx,cy,cw,ch,, ahk_id %A_LoopField%     ;get current controls pos
      If ( cx >= BoxX AND cy >= BoxY AND cx + cw <= BoxW AND cy + ch <= BoxH ){
        ControlsInBox .= "|" . A_LoopField
        If ("Tab" = Control(A_LoopField,"Type")){           ;when a tab is in selection
          ListOfCtrlOnTab := Tab_GetAllCtrlHWND(A_LoopField)          ;get all controls of that tab
          Loop, Parse, ListOfCtrlOnTab, |
          {                                                        ;and check if they are in box too
             ControlGetPos, cx,cy,cw,ch,, ahk_id %A_LoopField%
             If ( cx >= BoxX AND cy >= BoxY AND cx + cw <= BoxW AND cy + ch <= BoxH )
                ControlsInBox .= "|" . A_LoopField
          }
        }
      }
    }
  StringTrimLeft, ControlsInBox, ControlsInBox, 1
  Sort, ControlsInBox, U D|
  If InStr(ControlsInBox,"|")
      Selector_vMultiHWND(ControlsInBox)                  ;set multiselection
  Else If ControlsInBox
      Selector_vCtrlHWND(ControlsInBox)                   ;set single selection
  Else
      Return 0
}

;###############################################################################
;##################   Resize selector                                                       
;###############################################################################

;selector gui gets resized
Selector_Size:
  Selector_Size()
Return
Selector_Size(NewState=""){
  static State := True
  If NewState is not space    ;when called with state
    Return State := NewState         
  If !State
      Return
    
  If Selector_vCtrlHWND() {                  ;when single selection  
    If (GRID() AND !GetKeyState("LControl")){       ;when Grid and NOT LCtrl down
      State := False                                     ;disable itself to resize selector, without calling itself
      Selector_MoveCtrl2SelectorWGrid()                  ;update control in grid
      SetTimer, Selector_UpdateSet2PosAfterSize, 50      ;update selector to new control position when resize finished
      State := True                                      ;enable itself
    }
    else Selector_MoveCtrl2Selector()              ;otherwise move the control directly to follow the selector
  }Else If Selector_vMultiHWND()                ;when multi selection
    SetTimer, Selector_Multi2SelectorAfterSize, 10 ;update controls according to new selector pos/size when resize finished
  Else
    Selector_Set2PosSize(-10,-10)
  Grid_Draw()
  Settimer, ToolTip_ShowTips4Pos, -50
}

;update selector to control dimentions 
; called when: 
;   - after selector resize with GRID in single selection mode
;   - after selector move with GRID in single selection mode (Messages_OnMouseMove())
Selector_UpdateSet2PosAfterSize:
  If !GetKeyState("LButton"){
      SetTimer, Selector_UpdateSet2PosAfterSize, Off
      Selector_Set2Ctrl()                    ;update selector to control dimentions
      Grid_Draw()                            ;Redraw grid
  }
Return

;update controls according to new selector pos/size when resize finished
Selector_Multi2SelectorAfterSize:
  If !GetKeyState("LButton"){                           ;wait for LButton release, then resize finsihed
      SetTimer, Selector_Multi2SelectorAfterSize, Off        ;disable timer 
      Selector_MoveCtrl2Selector()                                  ;update controls to new selector position/size 
      Grid_Draw()                                            ;redraw grid 
  }
Return

;###############################################################################
;##################   Update control pos/size according to selectors pos/size                                                       
;###############################################################################

; Update control pos/size to match selector's size and position
; called when:
;   - selector resizes in single selection
;   - after selector resize in multi selection
;   - after mouse move with LButton down on selector  
Selector_MoveCtrl2Selector:
  Selector_MoveCtrl2Selector()
Return
Selector_MoveCtrl2Selector(){
  WinGetPos, ux, uy,,, % "ahk_id " Selector_vGuiHWND()        ;get pos of user gui
  WinGetPos, sx,sy,sw,sh, % "ahk_id " Selector_vHWND()        ;get current pos/size of selector
  If (CtrlHWND := Selector_vCtrlHWND())                       ;when single selection
    ControlMove,, sx-ux, sy-uy, sw, sh, ahk_id %CtrlHWND%            ;move control to selector pos/size
  Else If (ListCtrlHWND := Selector_vMultiHWND()){            ;when multi selection
    Selector_vPosSize(dx, dy, dw, dh)                                ;get selectors previous pso/size
    fx := sw / dw , fy := sh / dh                                    ;calculate factor for x/y
    dx := sx - dx + (fx - 1) * (ux - dx) , dy := sy - dy + (fy - 1) * (uy - dy) ;calculate offset for x/y 
    
    ;??? doesn't work well on resize: controls width/height are not adjusted at all
    ;changing controls position based on change in selector width/height doesn't work well too => truncation error 
    ;don't know how to do it better                                         
    ;need something like Anchor, I guess 
    ;maybe also something proportinal to width/hight, because smaller controls stay better in frame then larger once.                                   
    
    Loop, Parse, ListCtrlHWND, |                                 ;go through all controls in selection
      {                                           
        ControlGetPos, cx,cy,cw,ch,, ahk_id %A_LoopField%                   ;get current controls pos
        ControlMove,, cx * fx + dx, cy * fy + dy, , , ahk_id %A_LoopField%  ;move control to new pos
      }
    Selector_vPosSize(sx,sy,sw,sh)                               ;store new selector pos/size
  }
  SetTimer, Gui_Refresh, -100                                 ;refresh gui after a timeout
}

; Update single control pos/size to match selector's size and position when GRID is used
; called when:
;   - selector resizes
;   - after mouse move with LButton down on selector with "PositionOnly"
Selector_MoveCtrl2SelectorWGrid(Type="") {
  GuiHWND := Selector_vGuiHWND()
  WinGetPos, ux, uy, , , ahk_id %GuiHWND%                  ;get pos of user gui
  WinGetPos, cx, cy ,cw, ch, % "ahk_id " Selector_vHWND()       ;get current pos/size of selector

  If Gui(GuiHWND,"Toolwindow"){
    ;window frame height/width of a toolwindow gui
    SysGet, XFrame,  7   ;SM_CXFIXEDFRAME =  7   ;width of frame
    SysGet, YFrame,  8   ;SM_CYFIXEDFRAME =  8   ;height of frame
    SysGet, HCaption, 51 ;SM_CYSMCAPTION  = 51   ;height of caption, the whole title height is: SM_CYSMCAPTION + SM_CYFIXEDFRAME
  }Else{
    ;window frame height/width of a regular gui
    SysGet, XFrame, 32   ;SM_CXSIZEFRAME = 32    ;width of frame
    SysGet, YFrame, 33   ;SM_CYSIZEFRAME = 33    ;height of frame
    SysGet, HCaption, 4  ;SM_CYCAPTION   =  4    ;height of caption, the whole title height is: SM_CYCAPTION + SM_CYSIZEFRAME
  }

  cx -= ux, cy -= uy                                              ;calculate relative position
  GRID := Grid()                                                  ;get GRID size
  dx := Mod(cx - XFrame, GRID)                            ;calculate x/y with snap to grid
  dy := Mod(cy - YFrame - HCaption, GRID)             
  dx -= (dx ? GRID : 0),cx -= dx, dy -= (dy ? GRID : 0),cy -= dy
  If !(Type = "PositionOnly"){                                    ;when also w/h should be adjusted (on selector resize)
    dw := Mod(cw, GRID), dh := Mod(ch, GRID)                      ;calculate w/h with snap to grid 
    cw -= (cw > dw ? dw : 0), ch -= (ch > dh ? dh : 0)
  }
  ControlMove,,cx,cy,cw,ch, % "ahk_id " Selector_vCtrlHWND()      ;move control to new pos
}

;###############################################################################
;##################   Delete control(s) in selection                                                       
;###############################################################################

;context sensitive hotkey to delete selection got pressed
Selector_DeleteControl:
  Selector_DeleteControl()                          ;delete selected control(s)
  WinActivate, % "ahk_id " Selector_vGuiHWND()      ;hide selector (move behind) by activating selected gui
  Grid_Draw()                                       ;redraw grid when control(s) deleted
return
;delete selected controls 
Selector_DeleteControl(){
  ListCtrlHWND := Selector_vCtrlHWND() . Selector_vMultiHWND()
  Loop, Parse, ListCtrlHWND, |                     ;go through selected controls
    If ("Tab" = Control(A_LoopField,"Type"))          ;delete tab
      Tab_DeleteItself(A_LoopField)
    Else                                              ;delete all other controls
      Control_Delete(A_LoopField)
}

